{***********************************************************************

Copyright (C) 2016 Bergen A van <x1@ziggo.nl>

This program is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see <http://www.gnu.org/licenses/>.

***********************************************************************}

PROGRAM p3; { weather report explorer }

{$I+} {$R+}

USES

p3m1, { date                               }
p3m2, { element = p3m2_Tweatherreport      }
p3m3, { element = p3m3_Ttransaction        }
p3m4, { sequential file of elements (p3m2) }
p3m5, { sequential file of elements (p3m3) }
p3m6; { sequential file (p3m5) sort        }

CONST

Cmaxdigits=6; { decimal digits --> } Cmaxinteger=999999;

Cmaxstringlength  =80;
Cmaxargumentnumber=4;

Cmasterfilename   ='weatherreport.bin';
Coldmasterfilename='weatherreport.bin~';

TYPE

Tinteger=-Cmaxinteger..Cmaxinteger;

Tstringposition=0..Cmaxstringlength;
Targumentnumber=0..Cmaxargumentnumber;

Tstring  =STRING[Cmaxstringlength];
Tpathname=Tstring;
Toption  =Tstring;
Tmessage =Tstring;

Ttest =FUNCTION(x:p3m2_Tweatherreport):BOOLEAN;
Tapply=PROCEDURE(x:p3m2_Tweatherreport);

VAR

option         :Toption;
errorlist      :ARRAY[1..Cmaxargumentnumber] OF Tmessage;
errorcount     :Targumentnumber;
masterfile     :p3m4_Pfile;
inputmasterfile:p3m4_Pinputfile;
source         :Tpathname;
weatherreport  :p3m2_Tweatherreport;
key            :ARRAY[1..2] OF p3m2_Tweatherreportkey;
stationnumber  :p3m2_Tstationnumber;
temperature    :ARRAY[1..2] OF p3m2_Ttemperature;
test           :Ttest;
apply          :Tapply;

PROCEDURE Rhalt(msg:Tmessage);

  BEGIN
    WRITELN(msg);
    HALT
  END;

FUNCTION Rfilexists(filename:Tpathname):BOOLEAN;

  VAR f    :FILE;
      found:BOOLEAN;

  BEGIN
    IF filename=''
    THEN found:=FALSE
    ELSE BEGIN
           ASSIGN(f,filename);
           {$I-}
           RESET(f);
           {$I+}
           found:=(IORESULT=0);
           CLOSE(f)
         END;
    IF NOT found THEN WRITELN('File '''+filename+''' not found.');
    Rfilexists:=found
  END;

PROCEDURE Rerasefile(filename:Tpathname);

  VAR f:FILE;

  BEGIN
    ASSIGN(f,filename);
    ERASE(f)
  END;

PROCEDURE Rrenamefile(oldfilename,newfilename:Tpathname);

  VAR f:FILE;

  BEGIN
    ASSIGN(f,oldfilename);
    RENAME(f,newfilename)
  END;

FUNCTION Rinteger2string(v:Tinteger):Tstring;

  BEGIN
    STR(v,Rinteger2string)
  END;

PROCEDURE Rstring2integer(    s:Tstring;
                          VAR v:Tinteger;
                          VAR e:BOOLEAN);

  VAR i:Tstringposition;

  BEGIN
    e:=(s='');
    IF NOT e THEN
      BEGIN
        e:=NOT(s[1] IN ['-','+','0'..'9']) OR ((LENGTH(s)=1) AND (s[1] IN ['-','+']));
        i:=1;
        WHILE NOT e AND (i<LENGTH(s)) DO
          BEGIN
            i:=i+1;
            e:=NOT(s[i] IN ['0'..'9']);
          END;
        e:=e OR (i>Cmaxdigits+1) OR ((i=Cmaxdigits+1) AND NOT(s[1] IN ['-','+']));
        IF NOT e THEN VAL(s,v)
      END
  END;

PROCEDURE Rgetstationnumber(    s:Tstring;
                            VAR x:p3m2_Tstationnumber;
                            VAR e:BOOLEAN);

  VAR v:Tinteger;

  BEGIN
    Rstring2integer(s,v,e);
    e:=e OR (v<100) OR (v>999);
    IF NOT e THEN x:=v
  END;

PROCEDURE Rgetdate(    s:Tstring;
                   VAR x:p3m1_Tdate;
                   VAR e:BOOLEAN);

  BEGIN
    p3m1_Rstring2date(s,x,e)
    { if we can get the current date }
    { then e:=e OR p3m1_Rlessdate(currentdate,x) }
  END;

PROCEDURE Rgettemperature(    s:Tstring;
                          VAR x:p3m2_Ttemperature;
                          VAR e:BOOLEAN);

  VAR v:Tinteger;

  BEGIN
    Rstring2integer(s,v,e);
    e:=e OR (v<p3m2_Cmintemperature) OR (v>p3m2_Cmaxtemperature);
    IF NOT e THEN x:=v
  END;

PROCEDURE Rnewmasterfile;

  VAR l:Tstring;

  BEGIN
    WRITELN('WARNING!');
    WRITELN('========');
    WRITELN('This will create an empty master file '''+Cmasterfilename+'''.');
    WRITE('Are you sure? (Type uppercase yes): ');READLN(l);
    IF l='YES'
    THEN BEGIN
           masterfile:=NEW(p3m4_Pfile,Minitialize(Cmasterfilename,TRUE));
           DISPOSE(masterfile,Mfinalize)
         END
    ELSE WRITELN('Command failed: invalid argument')
  END;

PROCEDURE Rupdatemasterfile(batchfilename:Tpathname);

  CONST Creportfilename        ='updatereport.txt';
        Ctransactionfilename   ='transaction.bin';
        Coldtransactionfilename='transaction.bin~';

  VAR reportfile          :TEXT;
      outputmasterfile    :p3m4_Poutputfile;
      oldmasterfile       :p3m4_Pfile;
      inputoldmasterfile  :p3m4_Pinputfile;
      transactionfile     :p3m5_Pfile;
      inputtransactionfile:p3m5_Pinputfile;
      m1,m2               :p3m2_Tweatherreport;
      sentinelkey,k1,k2,k3:p3m2_Tweatherreportkey;
      transaction         :p3m3_Ttransaction;
      processed,invalid   :p3m3_Ttransactionnumber;
      choice              :Tstring;
      infile,ok           :BOOLEAN;

  PROCEDURE Rvalidatebatch(VAR ok:BOOLEAN);

    CONST Creportfilename='validationreport.txt';
          Cmaxfieldnumber=6;

    TYPE Tfieldnumber=0..Cmaxfieldnumber;

    VAR errorlist                  :ARRAY[1..Cmaxfieldnumber] OF Tmessage;
        errorcount                 :Tfieldnumber;
        batchfile,reportfile       :TEXT;
        outputtransactionfile      :p3m5_Poutputfile;
        transaction                :p3m3_Ttransaction;
        linenumber,examined,invalid:p3m3_Ttransactionnumber;
        line                       :Tstring;

    PROCEDURE Rwriteheading(VAR f:TEXT);

      BEGIN
        WRITELN(f,'Batch file '''+batchfilename+''':');
        WRITELN(f);
      END;

    PROCEDURE Rwritefooting(VAR f:TEXT);

      BEGIN
        WRITELN(f);
        WRITELN(f,' Total transactions examined = ',examined:Cmaxdigits);
        WRITELN(f,' Total transactions in error = ',invalid:Cmaxdigits);
        WRITELN(f,' Total correct transactions  = ',examined-invalid:Cmaxdigits)
      END;

    PROCEDURE Rrecorderror(msg:Tmessage);

      BEGIN
        errorcount:=errorcount+1;
        errorlist[errorcount]:=msg
      END;

    PROCEDURE Rwriterrors(VAR f:TEXT);

      VAR i:Tfieldnumber;

      BEGIN
        invalid:=invalid+1;
        WRITELN(f,' line ',linenumber,': '''+line+'''');
        FOR i:=1 TO errorcount DO WRITELN(f,'  '+errorlist[i]);
        WRITELN(f);
      END;

    PROCEDURE Rchecktransaction;

      VAR field:Tstring;

      FUNCTION Rfieldexists(n:Tfieldnumber):BOOLEAN;

        VAR found:BOOLEAN;
            i    :Tstringposition;
            a    :Tfieldnumber;

        BEGIN
          field:='';
          i:=0;a:=0;
          WHILE (i<LENGTH(line)) AND (a+1<n) DO
            BEGIN
              i:=i+1;
              IF line[i]=',' THEN a:=a+1
            END;
          WHILE (i<LENGTH(line)) AND (line[i+1]<>',') DO
            BEGIN
              i:=i+1;
              IF NOT ((field='') AND (line[i]=' ')) THEN field:=field+line[i]
            END;
          found:=(a+1=n);
          IF NOT found THEN Rrecorderror('field '+Rinteger2string(n)+' is missing');
          Rfieldexists:=found
        END;

      PROCEDURE Rcheckweatherreportkey;

        PROCEDURE Rcheckstationnumber;

          VAR n:Tfieldnumber;
              e:BOOLEAN;

          BEGIN
            n:=1;
            IF Rfieldexists(n) THEN
              BEGIN
                Rgetstationnumber(field,transaction.Fweatherreportkey.Fstationnumber,e);
                IF e THEN Rrecorderror('field '+Rinteger2string(n)+': invalid station number')
              END
          END;

        PROCEDURE Rcheckdate;

          VAR n:Tfieldnumber;
              e:BOOLEAN;

          BEGIN
            n:=2;
            IF Rfieldexists(n) THEN
              BEGIN
                Rgetdate(field,transaction.Fweatherreportkey.Fdate,e);
                e:=e OR NOT p3m2_Rlesskey(transaction.Fweatherreportkey,sentinelkey);
                IF e THEN Rrecorderror('field '+Rinteger2string(n)+': invalid date')
              END
          END;

        BEGIN { Rcheckweatherreportkey }
          Rcheckstationnumber;
          Rcheckdate
        END;

      PROCEDURE Rcheckevent;

        VAR n:Tfieldnumber;

        BEGIN
          n:=3;
          IF Rfieldexists(n) THEN
            CASE field OF
              'n': transaction.Fevent:=addition;
              'x': transaction.Fevent:=deletion
              ELSE Rrecorderror('field '+Rinteger2string(n)+': invalid event')
              END
        END;

      PROCEDURE Rcheckrest;

        PROCEDURE Rcheckelements;

          PROCEDURE Rchecktemperature(    n:Tfieldnumber;
                                      VAR x:p3m2_Ttemperature);

            VAR e:BOOLEAN;

            BEGIN
              IF Rfieldexists(n) THEN
                BEGIN
                  Rgettemperature(field,x,e);
                  IF e THEN Rrecorderror('field '+Rinteger2string(n)+': invalid temperature')
                END
            END;

          PROCEDURE Rcheckt1;

            BEGIN
              Rchecktemperature(4,transaction.Felement.Ft1)
            END;

          PROCEDURE Rcheckt2;

            BEGIN
              Rchecktemperature(5,transaction.Felement.Ft2);
            END;

          PROCEDURE Rcheckt3;

            BEGIN
              Rchecktemperature(6,transaction.Felement.Ft3);
            END;

          BEGIN { Rcheckelements }
            Rcheckt1;
            Rcheckt2;
            Rcheckt3
          END;

        BEGIN { Rcheckrest }
          CASE transaction.Fevent OF
            addition: Rcheckelements;
            deletion:
            END;
        END;

      BEGIN { Rchecktransaction }
        errorcount:=0;
        transaction.Ftransactionnumber:=linenumber;
        Rcheckweatherreportkey;
        Rcheckevent;
        Rcheckrest;
        IF errorcount>0
        THEN Rwriterrors(reportfile)
        ELSE outputtransactionfile^.Mwrite(transaction);
        examined:=examined+1;
      END;

    BEGIN { Rvalidatebatch }
      WRITE('Validating batch:');
      transactionfile:=NEW(p3m5_Pfile,Minitialize(Ctransactionfilename,TRUE));
      outputtransactionfile:=NEW(p3m5_Poutputfile,Minitialize(transactionfile,100));
      ASSIGN(batchfile,batchfilename);RESET(batchfile);
      ASSIGN(reportfile,Creportfilename);REWRITE(reportfile);
      Rwriteheading(reportfile);
      examined:=0;invalid:=0;
      linenumber:=0;
      WHILE NOT EOF(batchfile) DO
        BEGIN
          IF linenumber=p3m3_Cmaxtransactionnumber THEN Rhalt('error p3-1');
          linenumber:=linenumber+1;
          READLN(batchfile,line);
          Rchecktransaction
        END;
      Rwritefooting(reportfile);Rwritefooting(OUTPUT);
      CLOSE(reportfile);
      CLOSE(batchfile);
      DISPOSE(outputtransactionfile,Mfinalize);
      DISPOSE(transactionfile,Mfinalize);
      OK:=(invalid=0);
      IF NOT OK THEN WRITELN(' See '''+Creportfilename+''' for more details.')
    END;

  PROCEDURE Rsorttransactionfile;

    VAR oldtransactionfile:p3m5_Pfile;

    BEGIN
      Rrenamefile(Ctransactionfilename,Coldtransactionfilename);
      oldtransactionfile:=NEW(p3m5_Pfile,Minitialize(Coldtransactionfilename,FALSE));
      transactionfile:=NEW(p3m5_Pfile,Minitialize(Ctransactionfilename,TRUE));
      p3m6_Rfilesort('','','','',oldtransactionfile,transactionfile);
      DISPOSE(transactionfile,Mfinalize);
      DISPOSE(oldtransactionfile,Mfinalize);
      Rerasefile(Coldtransactionfilename)
    END;

  PROCEDURE Rwriteheading(VAR f:TEXT);

      BEGIN
        WRITELN(f,'Batch file '''+batchfilename+''':');
        WRITELN(f)
      END;

  PROCEDURE Rwritefooting(VAR f:TEXT);

      BEGIN
        WRITELN(f);
        WRITELN(f,' Total transactions processed = ',processed:Cmaxdigits);
        WRITELN(f,' Total transactions in error  = ',invalid:Cmaxdigits);
        WRITELN(f,' Total correct transactions   = ',processed-invalid:Cmaxdigits)
      END;

  PROCEDURE Rwriterror(VAR f  :TEXT;
                           msg:Tmessage);

    BEGIN
      invalid:=invalid+1;
      WRITELN(f,' line ',transaction.Ftransactionnumber:Cmaxdigits,': '+msg)
    END;

  PROCEDURE Rreadtransactionfile(VAR x:p3m2_Tweatherreportkey);

    BEGIN
      inputtransactionfile^.Mread(transaction);
      IF inputtransactionfile^.Meof
      THEN x:=sentinelkey
      ELSE x:=transaction.Fweatherreportkey
    END;

  PROCEDURE Rreadoldmasterfile(VAR x:p3m2_Tweatherreportkey);

    BEGIN
      inputoldmasterfile^.Mread(m1);
      IF inputoldmasterfile^.Meof
      THEN x:=sentinelkey
      ELSE p3m2_Rgetkey(m1,x)
    END;

  PROCEDURE Rchoosekey(VAR x:p3m2_Tweatherreportkey);

    BEGIN
      IF p3m2_Rlesskey(k1,k2)
      THEN x:=k1
      ELSE x:=k2
    END;

  PROCEDURE Rgetinitialstatus;

    BEGIN
      infile:=p3m2_Rsamekey(k3,k2);
      IF infile THEN
        BEGIN
          m2:=m1;
          Rreadoldmasterfile(k2)
        END
    END;

  PROCEDURE Rapplytransaction;

    BEGIN
      CASE transaction.Fevent OF
        addition:
          IF infile
          THEN Rwriterror(reportfile,'key already in master file')
          ELSE BEGIN
                 m2.Fstationnumber:=transaction.Fweatherreportkey.Fstationnumber;
                 m2.Fdate:=transaction.Fweatherreportkey.Fdate;
                 m2.Felement:=transaction.Felement;
                 infile:=TRUE
               END;
        deletion:
          IF NOT infile
          THEN Rwriterror(reportfile,'key not in master file')
          ELSE infile:=FALSE
        END;
      processed:=processed+1
    END;

  PROCEDURE Rrecordfinalstatus;

    BEGIN
      IF infile THEN outputmasterfile^.Mwrite(m2)
    END;

  BEGIN { Rupdatemasterfile }
    IF Rfilexists(batchfilename) AND Rfilexists(Cmasterfilename) THEN
      BEGIN
        sentinelkey.Fstationnumber:=999;
        p3m1_Rnumber2date(p3m1_Cmaxdatenumber,sentinelkey.Fdate);
        Rvalidatebatch(ok);
        IF OK THEN
          BEGIN
            WRITE('Update master file? (Type uppercase yes): ');READLN(choice);
            IF choice<>'YES'
            THEN WRITELN('Command failed: invalid argument')
            ELSE BEGIN
                   WRITE('Updating master file:');
                   Rsorttransactionfile;
                   Rrenamefile(Cmasterfilename,Coldmasterfilename);
                   oldmasterfile       :=NEW(p3m4_Pfile,Minitialize(Coldmasterfilename,FALSE));
                   inputoldmasterfile  :=NEW(p3m4_Pinputfile,Minitialize(oldmasterfile));
                   masterfile          :=NEW(p3m4_Pfile,Minitialize(Cmasterfilename,TRUE));
                   outputmasterfile    :=NEW(p3m4_Poutputfile,Minitialize(masterfile,100));
                   transactionfile     :=NEW(p3m5_Pfile,Minitialize(Ctransactionfilename,FALSE));
                   inputtransactionfile:=NEW(p3m5_Pinputfile,Minitialize(transactionfile));
                   ASSIGN(reportfile,Creportfilename);REWRITE(reportfile);
                   Rwriteheading(reportfile);
                   processed:=0;invalid:=0;
                   Rreadtransactionfile(k1);
                   Rreadoldmasterfile(k2);
                   Rchoosekey(k3);
                   WHILE NOT p3m2_Rsamekey(k3,sentinelkey) DO
                     BEGIN
                       Rgetinitialstatus;
                       WHILE p3m2_Rsamekey(k3,k1) DO
                         BEGIN
                           Rapplytransaction;
                           Rreadtransactionfile(k1)
                         END;
                       Rrecordfinalstatus;
                       Rchoosekey(k3)
                     END;
                   Rwritefooting(reportfile);Rwritefooting(OUTPUT);
                   CLOSE(reportfile);
                   DISPOSE(inputtransactionfile,Mfinalize);
                   DISPOSE(transactionfile,Mfinalize);
                   DISPOSE(outputmasterfile,Mfinalize);
                   DISPOSE(masterfile,Mfinalize);
                   DISPOSE(inputoldmasterfile,Mfinalize);
                   DISPOSE(oldmasterfile,Mfinalize);
                   IF invalid>0 THEN WRITELN(' See '''+Creportfilename+''' for more details.')
                 END
          END;
        Rerasefile(Ctransactionfilename)
      END
  END;

PROCEDURE Rwriteweatherreport(x:p3m2_Tweatherreport);

  VAR date:Tstring;

  BEGIN
    p3m1_Rdate2string(x.Fdate,date);
    WITH x,Felement DO WRITELN(Fstationnumber:3,',',date,',',Ft1:4,',',Ft2:4,',',Ft3:4);
  END;

PROCEDURE Rrecorderror(msg:Tmessage);

  BEGIN
    errorcount:=errorcount+1;
    errorlist[errorcount]:=msg
  END;

PROCEDURE Rwriterrors;

  VAR i:Targumentnumber;

  BEGIN
    FOR i:=1 TO errorcount DO WRITELN(errorlist[i])
  END;

FUNCTION Rargumentexists(n:Targumentnumber):BOOLEAN;

  VAR found:BOOLEAN;

  BEGIN
    found:=(1{option}+n<=PARAMCOUNT);
    IF NOT found THEN Rrecorderror('argument '+Rinteger2string(n)+' is missing');
    Rargumentexists:=found
  END;

PROCEDURE Rcheckpathname(    n:Targumentnumber;
                         VAR x:Tpathname);

  BEGIN
    IF Rargumentexists(n) THEN x:=PARAMSTR(1+n)
  END;

PROCEDURE Rcheckstationnumber(    n:Targumentnumber;
                              VAR x:p3m2_Tstationnumber);

  VAR e:BOOLEAN;

  BEGIN
    IF Rargumentexists(n) THEN
      BEGIN
        Rgetstationnumber(PARAMSTR(1+n),x,e);
        IF e THEN Rrecorderror('argument '+Rinteger2string(n)+': invalid station number')
      END
  END;

PROCEDURE Rcheckdate(    n:Targumentnumber;
                     VAR x:p3m1_Tdate);

  VAR e:BOOLEAN;

  BEGIN
    IF Rargumentexists(n) THEN
      BEGIN
        Rgetdate(PARAMSTR(1+n),x,e);
        IF e THEN Rrecorderror('argument '+Rinteger2string(n)+': invalid date')
      END
  END;

PROCEDURE Rchecktemperature(    n:Targumentnumber;
                            VAR x:p3m2_Ttemperature);

  VAR e:BOOLEAN;

  BEGIN
    IF Rargumentexists(n) THEN
      BEGIN
        Rgettemperature(PARAMSTR(1+n),x,e);
        IF e THEN Rrecorderror('argument '+Rinteger2string(n)+': invalid temperature')
      END
  END;

FUNCTION Rsamekey(x:p3m2_Tweatherreport):BOOLEAN;

  VAR y:p3m2_Tweatherreportkey;

  BEGIN
    p3m2_Rgetkey(x,y);
    Rsamekey:=p3m2_Rsamekey(y,key[1])
  END;

FUNCTION Rkeyinrange(x:p3m2_Tweatherreport):BOOLEAN;

  VAR y:p3m2_Tweatherreportkey;

  BEGIN
    p3m2_Rgetkey(x,y);
    Rkeyinrange:=NOT p3m2_Rlesskey(y,key[1]) AND NOT p3m2_Rlesskey(key[2],y)
  END;

FUNCTION Rt1inrange(x:p3m2_Tweatherreport):BOOLEAN;

  BEGIN
    Rt1inrange:=(x.Fstationnumber=stationnumber) AND (x.Felement.Ft1>=temperature[1]) AND (x.Felement.Ft1<=temperature[2])
  END;

FUNCTION Rt2inrange(x:p3m2_Tweatherreport):BOOLEAN;

  BEGIN
    Rt2inrange:=(x.Fstationnumber=stationnumber) AND (x.Felement.Ft2>=temperature[1]) AND (x.Felement.Ft2<=temperature[2])
  END;

FUNCTION Rt3inrange(x:p3m2_Tweatherreport):BOOLEAN;

  BEGIN
    Rt3inrange:=(x.Fstationnumber=stationnumber) AND (x.Felement.Ft3>=temperature[1]) AND (x.Felement.Ft3<=temperature[2])
  END;

PROCEDURE Rlocate(    test:Ttest;
                  VAR x   :p3m2_Tweatherreport);

 VAR found:BOOLEAN;

 BEGIN
   found:=FALSE;
   inputmasterfile^.Mread(x);
   WHILE NOT(inputmasterfile^.Meof OR found) DO
     BEGIN
       found:=test(x);
       IF NOT found THEN inputmasterfile^.Mread(x)
     END
 END;

PROCEDURE Rtraverse(test :Ttest;
                    apply:Tapply);

 VAR x:p3m2_Tweatherreport;

 BEGIN
   inputmasterfile^.Mread(x);
   WHILE NOT inputmasterfile^.Meof DO
     BEGIN
       IF test(x) THEN apply(x);
       inputmasterfile^.Mread(x)
     END
 END;

BEGIN
  IF PARAMCOUNT=0 THEN WRITELN('Try ''p3 --help'' for more information.') ELSE
    BEGIN
      option:=PARAMSTR(1);
      CASE option OF
        '--help':
          BEGIN
            WRITELN('Weather report explorer.');
            WRITELN('Usage: p3 [OPTION] [ARGUMENT]...');
            WRITELN('Options:');
            WRITELN(' -n            |new master file '''+Cmasterfilename+'''');
            WRITELN(' -u SOURCE     |update master file where SOURCE is a batch file of TRANSACTION');
            WRITELN('     --help    |show this message');
            WRITELN('     --version |show version number');
            WRITELN;
            WRITELN('     --key       KEY                             |query');
            WRITELN('     --key-range KEY KEY                         |query');
            WRITELN('     --t1-range  STATION TEMPERATURE TEMPERATURE |query');
            WRITELN('     --t2-range  STATION TEMPERATURE TEMPERATURE |query');
            WRITELN('     --t3-range  STATION TEMPERATURE TEMPERATURE |query');
            WRITELN;
            WRITELN('Query arguments:');
            WRITELN;
            WRITELN(' KEY         = STATION DATE');
            WRITELN(' STATION     = 100..999');
            WRITELN(' DATE        = yyyymmdd (yyyyy=year mm=month dd=day)');
            WRITELN(' TEMPERATURE = ',p3m2_Cmintemperature,'..',p3m2_Cmaxtemperature,' (in 0.1 degrees Celsius)');
            WRITELN;
            WRITELN('Serial-access master file update:');
            WRITELN;
            WRITELN(' master record          = ''STATION,DATE,t1,t2,t3''');
            WRITELN(' TRANSACTION (addition) = ''STATION,DATE,n,t1,t2,t3''');
            WRITELN(' TRANSACTION (deletion) = ''STATION,DATE,x''');
            WRITELN;
            WRITELN(' where');
            WRITELN;
            WRITELN('  t1 = mean day TEMPERATURE');
            WRITELN('  t2 = minimum day TEMPERATURE');
            WRITELN('  t3 = maximum day TEMPERATURE');
            WRITELN;
            WRITELN('Examples:');
            WRITELN;
            WRITELN(' p3 -u batch.txt');
            WRITELN(' p3 --key 280 20161230');
            WRITELN(' p3 --key-range 100 19000101 999 20161230 >result.txt');
            WRITELN(' p3 --t2-range 280 -300 -100');
            WRITELN(' p3 --t3-range 280 300 400');
          END;
        '--version':
          BEGIN
            WRITELN('P3 version 1.0.0');
            WRITELN('Copyright (C) 2016 Bergen A van <x1@ziggo.nl>');
            WRITELN('License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>.');
            WRITELN('This is free software: you are free to change and redistribute it.');
            WRITELN('There is NO WARRANTY, to the extent permitted by law.');
          END;
        '-n':Rnewmasterfile;
        '-u':
          BEGIN
            errorcount:=0;
            Rcheckpathname(1,source);
            IF errorcount>0
            THEN Rwriterrors
            ELSE Rupdatemasterfile(source)
          END;
        '--key':
          BEGIN
            errorcount:=0;
            Rcheckstationnumber(1,key[1].Fstationnumber);
            Rcheckdate(2,key[1].Fdate);
            IF errorcount>0
            THEN Rwriterrors
            ELSE IF Rfilexists(Cmasterfilename) THEN
                   BEGIN
                     masterfile:=NEW(p3m4_Pfile,Minitialize(Cmasterfilename,FALSE));
                     inputmasterfile:=NEW(p3m4_Pinputfile,Minitialize(masterfile));
                     test:=@Rsamekey;
                     apply:=@Rwriteweatherreport;
                     Rlocate(test,weatherreport);
                     IF NOT inputmasterfile^.Meof THEN apply(weatherreport);
                     DISPOSE(inputmasterfile,Mfinalize);
                     DISPOSE(masterfile,Mfinalize)
                   END
          END;
        '--key-range':
          BEGIN
            errorcount:=0;
            Rcheckstationnumber(1,key[1].Fstationnumber);
            Rcheckdate(2,key[1].Fdate);
            Rcheckstationnumber(3,key[2].Fstationnumber);
            Rcheckdate(4,key[2].Fdate);
            IF errorcount>0
            THEN Rwriterrors
            ELSE IF Rfilexists(Cmasterfilename) THEN
                   BEGIN
                     masterfile:=NEW(p3m4_Pfile,Minitialize(Cmasterfilename,FALSE));
                     inputmasterfile:=NEW(p3m4_Pinputfile,Minitialize(masterfile));
                     test:=@Rkeyinrange;
                     apply:=@Rwriteweatherreport;
                     Rlocate(test,weatherreport);
                     WHILE NOT inputmasterfile^.Meof AND test(weatherreport) DO
                       BEGIN
                         apply(weatherreport);
                         inputmasterfile^.Mread(weatherreport)
                       END;
                     DISPOSE(inputmasterfile,Mfinalize);
                     DISPOSE(masterfile,Mfinalize)
                   END
          END;
        '--t1-range','--t2-range','--t3-range':
          BEGIN
            errorcount:=0;
            Rcheckstationnumber(1,stationnumber);
            Rchecktemperature(2,temperature[1]);
            Rchecktemperature(3,temperature[2]);
            IF errorcount>0
            THEN Rwriterrors
            ELSE IF Rfilexists(Cmasterfilename) THEN
                   BEGIN
                     masterfile:=NEW(p3m4_Pfile,Minitialize(Cmasterfilename,FALSE));
                     inputmasterfile:=NEW(p3m4_Pinputfile,Minitialize(masterfile));
                     CASE option OF
                       '--t1-range':test:=@Rt1inrange;
                       '--t2-range':test:=@Rt2inrange;
                       '--t3-range':test:=@Rt3inrange;
                       END;
                     apply:=@Rwriteweatherreport;
                     Rtraverse(test,apply);
                     DISPOSE(inputmasterfile,Mfinalize);
                     DISPOSE(masterfile,Mfinalize)
                   END
          END
        ELSE WRITELN('p3: invalid option '''+option+'''')
      END
    END
END.
