{***********************************************************************

Copyright (C) 2016 Bergen A van <x1@ziggo.nl>

This file is part of P3.

P3 is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your
option) any later version.

P3 is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see <http://www.gnu.org/licenses/>.

***********************************************************************}

UNIT p3m1; { date }

{$I+} {$R+}

INTERFACE

CONST

p3m1_Cstartyear=1800;
p3m1_Cendyear  =2199;

p3m1_Cmaxdatenumber=146100; { 01/01/1800=1; 12/31/2199=146100 }

p3m1_Cmaxstringlength=80;

TYPE

p3m1_Tyear =p3m1_Cstartyear..p3m1_Cendyear;
p3m1_Tmonth=1..12;
p3m1_Tday  =1..31;

p3m1_Tdate=RECORD
             Fyear :p3m1_Tyear;
             Fmonth:p3m1_Tmonth;
             Fday  :p3m1_Tday
         END;

p3m1_Tdatenumber=0..p3m1_Cmaxdatenumber;

p3m1_Tstring=STRING[p3m1_Cmaxstringlength];

FUNCTION p3m1_Rvaliddate(x:p3m1_Tdate):BOOLEAN;
FUNCTION p3m1_Rsamedate(x1,x2:p3m1_Tdate):BOOLEAN;
FUNCTION p3m1_Rlessdate(x1,x2:p3m1_Tdate):BOOLEAN;
FUNCTION p3m1_Rdate2number(x:p3m1_Tdate):p3m1_Tdatenumber;

PROCEDURE p3m1_Rnumber2date(    n:p3m1_Tdatenumber;
                            VAR x:p3m1_Tdate);

PROCEDURE p3m1_Rstring2date(    s:p3m1_Tstring; { 'yyyymmdd' }
                            VAR x:p3m1_Tdate;
                            VAR e:BOOLEAN);

PROCEDURE p3m1_Rdate2string(    x:p3m1_Tdate;
                            VAR s:p3m1_Tstring); { 'yyyymmdd' }

{**********************************************************************}

IMPLEMENTATION

TYPE Tdayinmonth=p3m1_Tday;
     Tdayinyear =1..366;

FUNCTION Rlastdayinmonth(y:p3m1_Tyear;
                         m:p3m1_Tmonth):Tdayinmonth;

  VAR d:Tdayinmonth;

  BEGIN
    CASE m OF
       1: d:=31;
       2: IF (y MOD 4=0) THEN d:=29 ELSE d:=28;
       3: d:=31;
       4: d:=30;
       5: d:=31;
       6: d:=30;
       7: d:=31;
       8: d:=31;
       9: d:=30;
      10: d:=31;
      11: d:=30;
      12: d:=31
      END;
    Rlastdayinmonth:=d
  END;

FUNCTION Rlastdayinyear(y:p3m1_Tyear):Tdayinyear;

  BEGIN
    IF (y MOD 4=0)
    THEN Rlastdayinyear:=366
    ELSE Rlastdayinyear:=365
  END;

FUNCTION p3m1_Rvaliddate(x:p3m1_Tdate):BOOLEAN;

  BEGIN
    p3m1_Rvaliddate:=(x.Fday<=Rlastdayinmonth(x.Fyear,x.Fmonth))
  END;

FUNCTION p3m1_Rsamedate(x1,x2:p3m1_Tdate):BOOLEAN;

  BEGIN
    p3m1_Rsamedate:=(x1.Fyear=x2.Fyear) AND (x1.Fmonth=x2.Fmonth) AND (x1.Fday=x2.Fday)
  END;

FUNCTION p3m1_Rlessdate(x1,x2:p3m1_Tdate):BOOLEAN;

  BEGIN
    p3m1_Rlessdate:=(x1.Fyear<x2.Fyear) OR ((x1.Fyear=x2.Fyear) AND ((x1.Fmonth<x2.Fmonth) OR ((x1.Fmonth=x2.Fmonth) AND (x1.Fday<x2.Fday))))
  END;

FUNCTION p3m1_Rdate2number(x:p3m1_Tdate):p3m1_Tdatenumber;

  VAR a:0..p3m1_Cmaxdatenumber;
      y:p3m1_Tyear;
      m:p3m1_Tmonth;

  BEGIN
    a:=0;
    y:=p3m1_Cstartyear;
    WHILE y<x.Fyear DO
      BEGIN
        a:=a+Rlastdayinyear(y);
        y:=y+1
      END;
    m:=1;
    WHILE m<x.Fmonth DO
      BEGIN
        a:=a+Rlastdayinmonth(y,m);
        m:=m+1
      END;
    a:=a+x.Fday;
    p3m1_Rdate2number:=a
  END;

PROCEDURE p3m1_Rnumber2date(    n:p3m1_Tdatenumber;
                            VAR x:p3m1_Tdate);

  VAR a:0..p3m1_Cmaxdatenumber;
      y:p3m1_Tyear;
      m:p3m1_Tmonth;

  BEGIN
    a:=0;
    y:=p3m1_Cstartyear;
    WHILE a+Rlastdayinyear(y)<n DO
      BEGIN
        a:=a+Rlastdayinyear(y);
        y:=y+1
      END;
    m:=1;
    WHILE a+Rlastdayinmonth(y,m)<n DO
      BEGIN
        a:=a+Rlastdayinmonth(y,m);
        m:=m+1
      END;
    x.Fyear:=y;
    x.Fmonth:=m;
    x.Fday:=n-a
  END;

PROCEDURE p3m1_Rstring2date(    s:p3m1_Tstring; { 'yyyymmdd' }
                            VAR x:p3m1_Tdate;
                            VAR e:BOOLEAN);

  VAR   y:0..9999;
      m,d:0..99;
        i:0..8;

  BEGIN
    e:=(LENGTH(s)<>8);
    IF NOT e THEN
      BEGIN
        i:=0;
        REPEAT
          i:=i+1;
          e:=NOT (s[i] IN ['0'..'9'])
        UNTIL e OR (i=8);
        IF NOT e THEN
          BEGIN
            VAL(COPY(s,1,4),y);
            e:=(y<p3m1_Cstartyear) OR (y>p3m1_Cendyear);
            IF NOT e THEN
              BEGIN
                VAL(COPY(s,5,2),m);
                e:=(m<1) OR (m>12);
                IF NOT e THEN
                  BEGIN
                    VAL(COPY(s,7,2),d);
                    e:=(d<1) OR (d>31);
                    IF NOT e THEN
                      BEGIN
                        x.Fyear:=y;
                        x.Fmonth:=m;
                        x.Fday:=d;
                        e:=NOT p3m1_Rvaliddate(x)
                      END
                  END
              END
          END
      END
  END;

PROCEDURE p3m1_Rdate2string(    x:p3m1_Tdate;
                            VAR s:p3m1_Tstring); { 'yyyymmdd' }

  VAR   y:STRING[4];
      m,d:STRING[2];

  BEGIN
    STR(x.Fyear,y);
    CASE LENGTH(y) OF
      1: y:='000'+y;
      2: y:='00'+y;
      3: y:='0'+y
      END;
    STR(x.Fmonth,m);
    IF LENGTH(m)=1 THEN m:='0'+m;
    STR(x.Fday,d);
    IF LENGTH(d)=1 THEN d:='0'+d;
    s:=y+m+d
  END;

END.
