{***********************************************************************

Copyright (C) 2016 Bergen A van <x1@ziggo.nl>

This file is part of P3.

P3 is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your
option) any later version.

P3 is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see <http://www.gnu.org/licenses/>.

***********************************************************************}

UNIT p3m5; { sequential file of elements (p3m3) }

{$I+} {$R+}

INTERFACE

USES p3m3; { element = p3m3_Ttransaction }

CONST

p3m5_Cmaxfilesize =100000; { block/file }
p3m5_Cmaxblocksize=100; { record/block }

TYPE

p3m5_Tfilesize   =0..p3m5_Cmaxfilesize;
p3m5_Tblocknumber=0..p3m5_Cmaxfilesize+1;

p3m5_Tblocksize        =0..p3m5_Cmaxblocksize;
p3m5_Tblockrecordnumber=0..p3m5_Cmaxblocksize;

p3m5_Tblock=RECORD
              Frecord   :ARRAY[1..p3m5_Cmaxblocksize] OF p3m3_Telement;
              Fblocksize:p3m5_Tblocksize
            END;

p3m5_Tpathname=STRING[80];

p3m5_Tfilestatus=(opened,closed);

p3m5_Pfile=^p3m5_Tfile;
p3m5_Tfile=OBJECT

  PRIVATE

    Ffile      :FILE OF p3m5_Tblock;
    Fblock     :p3m5_Tblock;
    Ffilename  :p3m5_Tpathname;
    Ffilestatus:p3m5_Tfilestatus;

    PROPERTY Pblocksize:p3m5_Tblocksize READ Fblock.Fblocksize;

    FUNCTION Mfilesize:p3m5_Tfilesize;

    { Pfilestatus=closed }
    PROCEDURE Mopenfile;

    PROCEDURE Mclosefile;
    PROCEDURE Mclearfile;
    PROCEDURE Mclearblock;

    PROCEDURE Mreadblock(VAR n:p3m5_Tblocknumber);

    PROCEDURE Mwriteblock(n:p3m5_Tblocknumber);

    PROCEDURE Mreadrecord(    n:p3m5_Tblockrecordnumber;
                          VAR x:p3m3_Telement);

    PROCEDURE Mwriterecord(n:p3m5_Tblockrecordnumber;
                           x:p3m3_Telement);

    PROCEDURE Mreplacerecord(n:p3m5_Tblockrecordnumber;
                             x:p3m3_Telement);

    PROCEDURE Mdeleterecord(n:p3m5_Tblockrecordnumber);

  PUBLIC

    PROPERTY Pfilename  :p3m5_Tpathname   READ Ffilename;
    PROPERTY Pfilestatus:p3m5_Tfilestatus READ Ffilestatus;

    CONSTRUCTOR Minitialize(filename:p3m5_Tpathname;
                            newfile :BOOLEAN);

    { Pfilestatus=closed }
    DESTRUCTOR Mfinalize;

  END;

p3m5_Tpercentage=1..100;

p3m5_Poutputfile=^p3m5_Toutputfile;
p3m5_Toutputfile=OBJECT

  PRIVATE

    Ffile          :p3m5_Pfile;
    Fblocknumber   :p3m5_Tblocknumber;
    Fblockoccupancy:p3m5_Tblocksize;

  PUBLIC

    { f^.Pfilestatus=closed }
    CONSTRUCTOR Minitialize(f{ile}        :p3m5_Pfile;
                            blockoccupancy:p3m5_Tpercentage);

    DESTRUCTOR Mfinalize;

    PROCEDURE Mwrite(x:p3m3_Telement);

  END;

p3m5_Pinputfile=^p3m5_Tinputfile;
p3m5_Tinputfile=OBJECT

  PRIVATE

    Ffile         :p3m5_Pfile;
    Fblocknumber  :p3m5_Tblocknumber;
    Frecordnumber :p3m5_Tblockrecordnumber;
    Frecorddeleted:BOOLEAN;
    Fblockchanged :BOOLEAN;

    { NOT Msof AND NOT Meof }
    PROCEDURE Mcontent(VAR x:p3m3_Telement);

    { NOT Meof }
    PROCEDURE Mnext;

  PUBLIC

    { f^.Pfilestatus=closed }
    CONSTRUCTOR Minitialize(f{ile}:p3m5_Pfile);

    DESTRUCTOR Mfinalize;

    FUNCTION Msof:BOOLEAN;
    FUNCTION Meof:BOOLEAN;

    { NOT Meof }
    PROCEDURE Mread(VAR x:p3m3_Telement);

  END;

p3m5_Pinputoutputfile=^p3m5_Tinputoutputfile;
p3m5_Tinputoutputfile=OBJECT(p3m5_Tinputfile)

  PUBLIC

    { NOT Msof AND NOT Meof }
    PROCEDURE Mwrite(x:p3m3_Telement);

    { NOT Msof AND NOT Meof }
    PROCEDURE Mreplace(x:p3m3_Telement);

    { NOT Msof AND NOT Meof }
    PROCEDURE Mdelete;

  END;

{**********************************************************************}

IMPLEMENTATION

TYPE Tmessage=STRING[20];

PROCEDURE Rhalt(msg:Tmessage);

 BEGIN
   WRITELN(msg);
   HALT
 END;

CONSTRUCTOR p3m5_Tfile.Minitialize(filename:p3m5_Tpathname;
                                   newfile :BOOLEAN);

  BEGIN
    IF filename='' THEN Rhalt('error p3m5-1');
    Ffilename:=filename;
    ASSIGN(Ffile,Ffilename);
    IF newfile
    THEN REWRITE(Ffile)
    ELSE BEGIN
           {$I-}
           RESET(Ffile);
           {$I+}
           IF IORESULT<>0 THEN Rhalt('error p3m5-2')
         END;
    Ffilestatus:=closed
  END;

DESTRUCTOR p3m5_Tfile.Mfinalize;

  BEGIN
    IF Ffilestatus=opened THEN Rhalt('error p3m5-3');
    CLOSE(Ffile)
  END;

FUNCTION p3m5_Tfile.Mfilesize:p3m5_Tfilesize;

  BEGIN
    Mfilesize:=FILESIZE(Ffile)
  END;

PROCEDURE p3m5_Tfile.Mopenfile;

  BEGIN
    IF Ffilestatus=opened THEN Rhalt('error p3m5-4');
    Ffilestatus:=opened
  END;

PROCEDURE p3m5_Tfile.Mclosefile;

  BEGIN
    Ffilestatus:=closed
  END;

PROCEDURE p3m5_Tfile.Mclearfile;

  BEGIN
    REWRITE(Ffile)
  END;

PROCEDURE p3m5_Tfile.Mclearblock;

  BEGIN
    Fblock.Fblocksize:=0
  END;

PROCEDURE p3m5_Tfile.Mreadblock(VAR n:p3m5_Tblocknumber);

  VAR found:BOOLEAN;

  BEGIN
    REPEAT
      SEEK(Ffile,n-1);READ(Ffile,Fblock);
      found:=(Fblock.Fblocksize>0);
      IF NOT found THEN n:=n+1
    UNTIL found OR (n>FILESIZE(Ffile))
  END;

PROCEDURE p3m5_Tfile.Mwriteblock(n:p3m5_Tblocknumber);

  BEGIN
    SEEK(Ffile,n-1);WRITE(Ffile,Fblock)
  END;

PROCEDURE p3m5_Tfile.Mreadrecord(    n:p3m5_Tblockrecordnumber;
                                 VAR x:p3m3_Telement);

  BEGIN
    x:=Fblock.Frecord[n]
  END;

PROCEDURE p3m5_Tfile.Mwriterecord(n:p3m5_Tblockrecordnumber;
                                  x:p3m3_Telement);

  VAR i:p3m5_Tblockrecordnumber;

  BEGIN
    WITH Fblock DO
      BEGIN
        Fblocksize:=Fblocksize+1;
        IF n<Fblocksize THEN FOR i:=Fblocksize DOWNTO n+1 DO Frecord[i]:=Frecord[i-1];
        Frecord[n]:=x
      END
  END;

PROCEDURE p3m5_Tfile.Mreplacerecord(n:p3m5_Tblockrecordnumber;
                                    x:p3m3_Telement);

  BEGIN
    Fblock.Frecord[n]:=x
  END;

PROCEDURE p3m5_Tfile.Mdeleterecord(n:p3m5_Tblockrecordnumber);

  VAR i:p3m5_Tblockrecordnumber;

  BEGIN
    WITH Fblock DO
      BEGIN
        IF n<Fblocksize THEN FOR i:=n TO Fblocksize-1 DO Frecord[i]:=Frecord[i+1];
        Fblocksize:=Fblocksize-1
      END
  END;

CONSTRUCTOR p3m5_Toutputfile.Minitialize(f{ile}        :p3m5_Pfile;
                                         blockoccupancy:p3m5_Tpercentage);

  BEGIN
    Ffile:=f{ile};
    Ffile^.Mopenfile;
    Ffile^.Mclearfile;
    Ffile^.Mclearblock;
    Fblocknumber:=1;
    Fblockoccupancy:=p3m5_Cmaxblocksize*blockoccupancy{%} DIV 100; {record/block}
    IF Fblockoccupancy=0 THEN Fblockoccupancy:=1
  END;

DESTRUCTOR p3m5_Toutputfile.Mfinalize;

  BEGIN
    IF Ffile^.Pblocksize>0 THEN Ffile^.Mwriteblock(Fblocknumber);
    Ffile^.Mclosefile
  END;

PROCEDURE p3m5_Toutputfile.Mwrite(x:p3m3_Telement);

  BEGIN
    IF Ffile^.Pblocksize=Fblockoccupancy THEN
      BEGIN
        Ffile^.Mwriteblock(Fblocknumber);
        Fblocknumber:=Fblocknumber+1;
        IF Fblocknumber>p3m5_Cmaxfilesize THEN Rhalt('error p3m5-5');
        Ffile^.Mclearblock
      END;
    Ffile^.Mwriterecord(Ffile^.Pblocksize+1,x)
  END;

CONSTRUCTOR p3m5_Tinputfile.Minitialize(f{ile}:p3m5_Pfile);

  BEGIN
    Ffile:=f{ile};
    Ffile^.Mopenfile;
    Ffile^.Mclearblock;
    Fblocknumber:=0;
    Frecordnumber:=0;
    Frecorddeleted:=FALSE;
    Fblockchanged:=FALSE
  END;

DESTRUCTOR p3m5_Tinputfile.Mfinalize;

  BEGIN
    IF Fblockchanged THEN Ffile^.Mwriteblock(Fblocknumber);
    Ffile^.Mclosefile
  END;

FUNCTION p3m5_Tinputfile.Msof:BOOLEAN;

  BEGIN
    Msof:=(Fblocknumber=0)
  END;

FUNCTION p3m5_Tinputfile.Meof:BOOLEAN;

  BEGIN
    Meof:=(Fblocknumber>0) AND (Fblocknumber>Ffile^.Mfilesize)
  END;

PROCEDURE p3m5_Tinputfile.Mnext;

  BEGIN
    IF Fblocknumber>Ffile^.Mfilesize THEN Rhalt('error p3m5-6');
    IF Frecordnumber>=Ffile^.Pblocksize
    THEN BEGIN
           IF Fblockchanged THEN
             BEGIN
               Ffile^.Mwriteblock(Fblocknumber);
               Fblockchanged:=FALSE
             END;
           Fblocknumber:=Fblocknumber+1;
           IF Fblocknumber<=Ffile^.Mfilesize THEN Ffile^.Mreadblock(Fblocknumber);
           Frecordnumber:=1
         END
    ELSE Frecordnumber:=Frecordnumber+1
  END;

PROCEDURE p3m5_Tinputfile.Mcontent(VAR x:p3m3_Telement);

  BEGIN
    IF Fblocknumber=0 THEN Rhalt('error p3m5-7');
    IF Fblocknumber>Ffile^.Mfilesize THEN Rhalt('error p3m5-8');
    Ffile^.Mreadrecord(Frecordnumber,x)
  END;

PROCEDURE p3m5_Tinputfile.Mread(VAR x:p3m3_Telement);

  BEGIN
    IF NOT Frecorddeleted THEN Mnext ELSE Frecorddeleted:=FALSE;
    IF NOT Meof THEN Mcontent(x)
  END;

PROCEDURE p3m5_Tinputoutputfile.Mwrite(x:p3m3_Telement);

  BEGIN
    IF Fblocknumber=0 THEN Rhalt('error p3m5-9');
    IF Fblocknumber>Ffile^.Mfilesize THEN Rhalt('error p3m5-10');
    IF Ffile^.Pblocksize=p3m5_Cmaxblocksize THEN Rhalt('error p3m5-11');
    Ffile^.Mwriterecord(Frecordnumber,x);
    Fblockchanged:=TRUE
  END;

PROCEDURE p3m5_Tinputoutputfile.Mreplace(x:p3m3_Telement);

  BEGIN
    IF Fblocknumber=0 THEN Rhalt('error p3m5-12');
    IF Fblocknumber>Ffile^.Mfilesize THEN Rhalt('error p3m5-13');
    Ffile^.Mreplacerecord(Frecordnumber,x);
    Fblockchanged:=TRUE
  END;

PROCEDURE p3m5_Tinputoutputfile.Mdelete;

  BEGIN
    IF Fblocknumber=0 THEN Rhalt('error p3m5-14');
    IF Fblocknumber>Ffile^.Mfilesize THEN Rhalt('error p3m5-15');
    Ffile^.Mdeleterecord(Frecordnumber);
    Frecorddeleted:=TRUE;
    Fblockchanged:=TRUE;
    IF Frecordnumber>Ffile^.Pblocksize THEN Mnext
  END;

END.
